/**
 *
 * AiCiA - Android IRC Client
 *
 * Copyright (C)2010 GORRY.
 *
 */

package net.gorry.aicia;

import java.util.ArrayList;
import java.util.concurrent.CountDownLatch;

import android.app.Notification;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.SharedPreferences;
import android.media.AudioManager;
import android.media.SoundPool;
import android.net.Uri;
import android.os.Bundle;
import android.os.Handler;
import android.os.IBinder;
import android.os.RemoteException;
import android.os.SystemClock;
import android.os.Vibrator;
import android.text.SpannableStringBuilder;
import android.util.Log;

/**
 *
 * IRCサービス
 *
 * @author GORRY
 *
 */
public class IRCService extends ForegroundService {
	private static final String TAG = "IRCService";
	private static final boolean V = false;
	private static final boolean D = false;
	private static final boolean I = true;

	/** IRCServiceのアクション名 */
	public static final String ACTION = "IRCService";

	/** IRCServiceの機能番号 */
	public static final int RECEIVE_MESSAGE = 1;
	/** */
	public static final int CREATE_SERVER = 2;
	/** */
	public static final int REMOVE_SERVER = 3;
	/** */
	public static final int CREATE_CHANNEL = 4;
	/** */
	public static final int REMOVE_CHANNEL = 5;
	/** */
	public static final int CHANGE_NICK = 6;
	/** */
	public static final int CHANGE_TOPIC = 7;
	/** */
	public static final int PING_PONG = 8;

	private static final int NOTIFY_SYSTEM = 0;
	private static final int NOTIFY_LOWMEMORY = 1;
	private static final int NOTIFY_SERVER_MESSAGE = 10000;

	private IRCServerList ircServerList = null;
	private Context me = null;

	private NotificationManager nm;
	private String mInputBox = "";
	private int mInputBoxSelStart = 0;
	private int mInputBoxSelEnd = 0;
	private boolean mDonate = false;
	private boolean mPrintDonate = false;
	private final Handler mHandler = new Handler();
	private boolean mSetAlarm = false;
	private boolean mSetReceiverHome = false;
    private HomeReceiver mHomeReceiver = new HomeReceiver();

	private SoundPool mSoundPool = null;
	private int mSound = 0;
	private int mSoundId = 0;
	private float mSoundVolume = 0.0F;

	private static final int mInputHistorySize = 20;
	private ArrayList<String> mInputHistory = new ArrayList<String>();

	/**
	 * コンストラクタ
	 */
	public IRCService() {
		super(NOTIFY_SYSTEM);
		me = this;
	}

	/*
	 * 作成
	 * @see android.app.Service#onCreate()
	 */
	@Override
	public void onCreate() {
		super.onCreate();
		if (I) Log.i(TAG, "onCreate()");
		nm = super.getNotificationManager();

		SystemConfig.setContext(me);
		SystemConfig.loadConfig();

		if (ircServerList == null) {
			ircServerList = new IRCServerList(me, true);
			ircServerList.addEventListener(new IRCServerListListener());
			ircServerList.reloadList();
			ircServerList.restoreServerIsConnectedFlag();
		}

		// startTimer();
		if (!mSetAlarm) {
			MyAlarmManager.setAlarmManager(this, new Runnable() {
				public void run() {
					if (D) Log.d(TAG, "receive AlarmManager");
					sendPingToMain();
				}
			});
			mSetAlarm = true;
		}

		if (!mSetReceiverHome) {
            IntentFilter iFilter = new IntentFilter();
            iFilter.addAction(Intent.ACTION_CLOSE_SYSTEM_DIALOGS);
            registerReceiver(mHomeReceiver, iFilter);
			mSetReceiverHome = true;
		}

		loadRingSound();

}

	/*
	 * 開始
	 * @see android.app.Service#onStart(android.content.Intent, int)
	 */
	@Override
	public void onStart(final Intent intent, final int startId) {
		if (I) Log.i(TAG, "onStart()");
		final Bundle extras = intent.getExtras();
		if (extras != null) {
			mDonate = (extras.getBoolean("donate"));
		}
		SharedPreferences pref = getApplicationContext().getSharedPreferences("ircservice", 0);
		final boolean killedByLowMemory = pref.getBoolean("killedbylowmemory", false);
		if (killedByLowMemory) {
			try {
				// ここで例外発生により落ちる事例がいくつかあるので防止。原因不明
				startForegroundCompat(
						showNotification(
								"Restart", (mDonate ? "AiCiA (DONATED)" : "AiCiA"),
								getString(R.string.ircservice_java_restartircservicebylowmemory),
								R.drawable.icon,
								(SystemConfig.fixSystemIcon ? Notification.FLAG_ONGOING_EVENT : 0),
								NOTIFY_SYSTEM
						)
				);
			} catch (final Exception e) {
				// e.printStackTrace();
			}
			final SharedPreferences.Editor editor = pref.edit();
			editor.putBoolean("killedbylowmemory", false);
			editor.commit();
		} else {
			try {
				// ここで例外発生により落ちる事例がいくつかあるので防止。原因不明
				startForegroundCompat(
						showNotification(
								"Start", (mDonate ? "AiCiA (DONATED)" : "AiCiA"),
								getString(R.string.ircservice_java_startircservice),
								R.drawable.icon,
								(SystemConfig.fixSystemIcon ? Notification.FLAG_ONGOING_EVENT : 0),
								NOTIFY_SYSTEM
						)
				);
			} catch (final Exception e) {
				// e.printStackTrace();
			}
		}
		clearNotification(NOTIFY_LOWMEMORY);
		pref = null;
	}

	private void sendPingToMain() {
		if (D) Log.d(TAG, "sendPingToMain()");
		
		// ActivityMainにKeepalive Pingを送る
		final Intent intent2 = new Intent(ACTION);
		intent2.putExtra("msg", IRCService.PING_PONG);
		sendBroadcast(intent2);
	}

	/**
	 * HOMEボタン押下を受け取るレシーバ
	 */
	public class HomeReceiver extends BroadcastReceiver {
	    @Override
	    public void onReceive(Context context, Intent intent) {
			if (D) Log.d(TAG, "receive HomeReceiver");
	    	sendPingToMain();
	    }
	}

	
	/*
	 * 破棄
	 * @see android.app.Service#onDestroy()
	 */
	@Override
	public void onDestroy() {
		super.onDestroy();

		if (I) Log.i(TAG, "onDestroy()");
		showNotification(
				"End", (mDonate ? "AiCiA (DONATED)" : "AiCiA"),
				getString(R.string.ircservice_java_shutdownircservice),
				R.drawable.icon,
				(SystemConfig.fixSystemIcon ? Notification.FLAG_ONGOING_EVENT : 0),
				NOTIFY_SYSTEM
		);
		if (ircServerList != null) {
			ircServerList.closeAll();
		}
		SharedPreferences pref = getApplicationContext().getSharedPreferences("ircservice", 0);
		final SharedPreferences.Editor editor = pref.edit();
		editor.putBoolean("killedbylowmemory", false);
		editor.commit();
		pref = null;
		clearNotification(NOTIFY_SYSTEM);
		clearNotification(NOTIFY_LOWMEMORY);
		for (int i=0; i<ircServerList.getNumServers(); i++) {
			clearNotification(NOTIFY_SERVER_MESSAGE+i);
		}
		// stopTimer();
        unregisterReceiver(mHomeReceiver);
		mSetReceiverHome = false;
		MyAlarmManager.resetAlarmManager(this);
		mSetAlarm = false;
		ircServerList.dispose();
		ircServerList = null;
	}

	// バインド
	/*
	 * バインド
	 * @see android.app.Service#onBind(android.content.Intent)
	 */
	@Override
	public IBinder onBind(final Intent intent) {
		if (I) Log.i(TAG, "onBind()");
//		me = this;
		final Bundle extras = intent.getExtras();
		if (extras != null) {
			mDonate = (extras.getBoolean("donate"));
		}

		/*
		SystemConfig.setContext(me);
		SystemConfig.loadConfig();

		if (ircServerList == null) {
			ircServerList = new IRCServerList(me, true);
			ircServerList.addEventListener(new IRCServerListListener());
			ircServerList.reloadList();
			ircServerList.restoreServerIsConnectedFlag();
		}
		*/
		
		if (!mDonate) {
			if (!mPrintDonate) {
				mPrintDonate = true;
				ircServerList.echoMessageToOther("You can select [AiCiA Donate Version]: https://market.android.com/details?id=net.gorry.aicia_donate");
			}
		}

		if (!IIRCService.class.getName().equals(intent.getAction())) {
			return null;
		}

		return ircServiceIf;
	}

	/*
	 * 再バインド
	 * @see android.app.Service#onRebind(android.content.Intent)
	 */
	@Override
	public void onRebind(final Intent intent) {
		if (I) Log.i(TAG, "onRebind()");
	}

	/*
	 * バインド解除
	 * @see android.app.Service#onUnbind(android.content.Intent)
	 */
	@Override
	public boolean onUnbind(final Intent intent) {
		if (I) Log.i(TAG, "onUnbind()");
		return true;
	}

	@Override
	public void onLowMemory() {
		if (I) Log.i(TAG, "onLowMemory()");
		showNotification(
				"Low Memory", (mDonate ? "AiCiA (DONATED)" : "AiCiA"),
				getString(R.string.ircservice_java_onlowmemory),
				R.drawable.icon_warn,
				0,
				NOTIFY_LOWMEMORY
		);
		SharedPreferences pref = getApplicationContext().getSharedPreferences("ircservice", 0);
		final SharedPreferences.Editor editor = pref.edit();
		editor.putBoolean("killedbylowmemory", true);
		editor.commit();
		pref = null;
		return;
	}

	//
	/**
	 * サーバリストイベントリスナ
	 */
	private class IRCServerListListener implements IRCServerListEventListener {
		public void receiveMessageToChannel(final String serverName, final String channel, final String nick, final String dateMsg, final SpannableStringBuilder ssb, final SpannableStringBuilder ssbOther, final boolean alert) {
			if (D) Log.d(TAG, "receiveMessageToChannel()");
			if (alert) {
				doAlert();
			}
			int serverId = ircServerList.getServerId(serverName);
			final IRCServer ircServer = ircServerList.getServer(serverName);
			int channelId = 0;
			if (ircServer != null) {
				channelId = ircServer.getChannelId(channel);
			}
			final Intent intent = new Intent(ACTION);
			intent.putExtra("msg", IRCService.RECEIVE_MESSAGE);
			intent.putExtra("server", serverName);
			intent.putExtra("channel", channel);
			intent.putExtra("nick", nick);
			intent.putExtra("date", dateMsg);
			intent.putExtra("ssb", ssb);
			intent.putExtra("ssbOther", ssbOther);
			intent.putExtra("alert", alert);
			intent.putExtra("serverid", serverId);
			intent.putExtra("channelid", channelId);
			sendBroadcast(intent);
		}

		public void createServer(final String serverName, final String dateMsg) {
			if (I) Log.i(TAG, "createServer()");
			ircServerList.setCurrentServerName(serverName);
			final Intent intent = new Intent(ACTION);
			intent.putExtra("msg", IRCService.CREATE_SERVER);
			intent.putExtra("server", serverName);
			intent.putExtra("date", dateMsg);
			sendBroadcast(intent);
		}

		public void removeServer(final String serverName, final String dateMsg) {
			if (I) Log.i(TAG, "removeServer()");
			final Intent intent = new Intent(ACTION);
			intent.putExtra("msg", IRCService.REMOVE_SERVER);
			intent.putExtra("server", serverName);
			intent.putExtra("date", dateMsg);
			sendBroadcast(intent);
		}

		public void createChannel(final String serverName, final String channel, final String dateMsg) {
			if (I) Log.i(TAG, "createChannel()");
			ircServerList.setCurrentServerName(serverName);
			ircServerList.setCurrentChannel(serverName, channel);
			final Intent intent = new Intent(ACTION);
			intent.putExtra("msg", IRCService.CREATE_CHANNEL);
			intent.putExtra("server", serverName);
			intent.putExtra("channel", channel);
			intent.putExtra("date", dateMsg);
			sendBroadcast(intent);
		}

		public void removeChannel(final String serverName, final String channel, final String dateMsg) {
			if (I) Log.i(TAG, "removeChannel()");
			final Intent intent = new Intent(ACTION);
			intent.putExtra("msg", IRCService.REMOVE_CHANNEL);
			intent.putExtra("server", serverName);
			intent.putExtra("channel", channel);
			intent.putExtra("date", dateMsg);
			sendBroadcast(intent);
		}

		public void changeNick(final String serverName, final String oldNick, final String newNick, final String dateMsg) {
			if (I) Log.i(TAG, "changeNick()");
			final Intent intent = new Intent(ACTION);
			intent.putExtra("msg", IRCService.CHANGE_NICK);
			intent.putExtra("server", serverName);
			intent.putExtra("oldnick", oldNick);
			intent.putExtra("newnick", newNick);
			intent.putExtra("date", dateMsg);
			sendBroadcast(intent);
		}

		public void changeTopic(final String serverName, final String channel, final String topic, final String dateMsg) {
			if (I) Log.i(TAG, "changeTopic()");
			final Intent intent = new Intent(ACTION);
			intent.putExtra("msg", IRCService.CHANGE_TOPIC);
			intent.putExtra("server", serverName);
			intent.putExtra("channel", channel);
			intent.putExtra("topic", topic);
			intent.putExtra("date", dateMsg);
			sendBroadcast(intent);
		}

		public void receiveConnect(final String serverName) {
			if (I) Log.i(TAG, "receiveConnect()");
			if (D) ircServerList.echoMessageToChannel(serverName, IRCMsg.sSystemChannelName, "receiveConnect()");
			if (SystemConfig.showServerIcon) {
				showNotification("Connect " + serverName, (mDonate ? "AiCiA (DONATED)" : "AiCiA"),
						String.format(me.getString(R.string.ircservice_java_connectserver),serverName),
						R.drawable.icon_server,
						0,
						NOTIFY_SERVER_MESSAGE + ircServerList.getServerId(serverName)
				);
			}
		}

		public void receiveDisconnect(final String serverName) {
			if (I) Log.i(TAG, "receiveDisconnect()");
			if (D) ircServerList.echoMessageToChannel(serverName, IRCMsg.sSystemChannelName, "receiveDisconnect()");
			if (SystemConfig.showServerIcon) {
				showNotification("Disconnect " + serverName, (mDonate ? "AiCiA (DONATED)" : "AiCiA"),
						String.format(me.getString(R.string.ircservice_java_disconnectserver),serverName),
						R.drawable.icon_disconnect,
						0,
						NOTIFY_SERVER_MESSAGE + ircServerList.getServerId(serverName)
				);
			}
			if (ircServerList.getNeedReconnect(serverName)) {
				ircServerList.echoMessageToChannel(serverName, IRCMsg.sSystemChannelName, "Wait " + SystemConfig.autoReconnectWaitSec + " seconds for retry.");
			}
		}
	}


	/**
	 * 通知表示
	 * @param ticker ティッカー
	 * @param title タイトル
	 * @param message メッセージ
	 * @param id ID
	 */
	private Notification showNotification(final String ticker, final String title, final String message, final int icon, final int flag, final int id) {
		if (I) Log.i(TAG, "showNotification(): id="+id+", ticker="+ticker+", title="+title+", message="+message);
		final Intent intent = new Intent(this, net.gorry.aicia.ActivityMain.class);
		if (id == 1) {
			//
		} else if (id >= NOTIFY_SERVER_MESSAGE) {
			final int serverId = id - NOTIFY_SERVER_MESSAGE;
			if (serverId < ircServerList.getNumServers()) {
				intent.putExtra("serverName", ircServerList.getServerName(serverId));
				intent.setData((Uri.parse("foobar://"+SystemClock.elapsedRealtime())));
			}
		}
		final Notification notification = new Notification(icon, ticker, System.currentTimeMillis());
		final PendingIntent pi = PendingIntent.getActivity(this, 0, intent, 0);
		notification.setLatestEventInfo(this, title, message, pi);
		notification.flags = flag;
		nm.cancel(id);
		nm.notify(id, notification);
		return notification;
	}

	/**
	 * 通知表示の消去
	 * @param ticker ティッカー
	 * @param title タイトル
	 * @param message メッセージ
	 * @param id ID
	 */
	private void clearNotification(final int id) {
		if (I) Log.i(TAG, "clearNotification(): id="+id);
		nm.cancel(id);
	}

	/**
	 * フラット化したサーバ名・チャンネルリストの取得
	 * @param flatServerName
	 * @param flatChannel
	 * @param flatChannelUpdated
	 * @param flatChannelAlerted
	 * @return
	 */
	private int getFlatServerChannelList(final ArrayList<String> flatServerName, final ArrayList<String> flatChannel, final ArrayList<Boolean> flatChannelUpdated, final ArrayList<Boolean> flatChannelAlerted ) {
		if (D) Log.d(TAG, "getFlatServerChannelList()");
		int noFlat = 0;
		int select = -1;
		final String currentServerName = ircServerList.getCurrentServerName();
		if (currentServerName == null) return -1;
		final String currentChannel = ircServerList.getCurrentChannel(currentServerName);
		if (currentChannel == null) return -1;
		final String[] svList = ircServerList.getServerList();
		for (int i=0; i<svList.length; i++) {
			final String[] chList = ircServerList.getChannelList(svList[i]);
			final Boolean[] chuList = ircServerList.getChannelUpdatedList(svList[i]);
			final Boolean[] chaList = ircServerList.getChannelAlertedList(svList[i]);
			for (int j=0; j<chList.length; j++) {
				flatServerName.add(svList[i]);
				flatChannel.add(chList[j]);
				flatChannelUpdated.add(chuList[j]);
				flatChannelAlerted.add(chaList[j]);
				if (svList[i].equalsIgnoreCase(currentServerName)
						&& chList[j].equalsIgnoreCase(currentChannel)) {
					select = noFlat;
				}
				noFlat++;
			}
		}
		return select;
	}

	/**
	 * IRCサービスAPI
	 */
	private final IIRCService.Stub ircServiceIf = new IIRCService.Stub() {
		public void shutdown() {
			if (I) Log.i(TAG, "shutdown()");
			mSetReceiverHome = false;
			MyAlarmManager.resetAlarmManager(me);
			mSetAlarm = false;
			ircServerList.closeAll();
			stopSelf();
		}

		public void reloadSystemConfig() {
			if (I) Log.i(TAG, "reloadSystemConfig()");
			SystemConfig.loadConfig();

			loadRingSound();
		}

		public CharSequence getSpanChannelLog(final String serverName, final String channel) throws RemoteException {
			if (I) Log.i(TAG, "getSpanChannelLog(): serverName="+serverName+", channel="+channel);
			return ircServerList.getSpanChannelLog(serverName, channel);
		}

		public void clearChannelLog(final String serverName, final String channel) throws RemoteException {
			if (I) Log.i(TAG, "clearChannelLog(): serverName="+serverName+", channel="+channel);
			ircServerList.clearChannelLog(serverName, channel);
		}

		public void clearAllChannelLog() throws RemoteException {
			if (I) Log.i(TAG, "clearAllChannelLog()");
			final int num = ircServerList.getNumServers();
			for (int i=0; i<num; i++) {
				final String serverName = ircServerList.getServerName(i);
				ircServerList.clearAllChannelLog(serverName);
			}
			ircServerList.clearAllChannelLog(IRCMsg.sOtherChannelName);
			ircServerList.echoMessageToOther("Sub-log was cleared.");
		}

		public void reloadChannelConfig(final String serverName, final String channel) throws RemoteException {
			if (I) Log.i(TAG, "reloadChannelConfig(): serverName="+serverName+", channel="+channel);
			ircServerList.getServer(serverName).loadChannelConfig(channel);
		}

		public String getNick(final String serverName) throws RemoteException {
			if (D) Log.d(TAG, "getNick(): serverName="+serverName);
			return ircServerList.getNick(serverName);
		}

		public String getTopic(final String serverName, final String channel) throws RemoteException {
			if (I) Log.i(TAG, "getTopic(): serverName="+serverName+", channel="+channel);
			return ircServerList.getTopic(serverName, channel);
		}

		public void sendCommandLine(final String serverName, final String channel, final String message) throws RemoteException {
			if (I) Log.i(TAG, "sendCommandLine(): : serverName="+serverName+", channel="+channel+", message="+message);
			ircServerList.sendCommandLine(serverName, channel, message);
		}

		public void sendMessageToChannel(final String serverName, final String channel, final String message) throws RemoteException {
			if (I) Log.i(TAG, "sendMessageToChannel(): serverName="+serverName+", channel="+channel+", message="+message);
			ircServerList.sendMessageToChannel(serverName, channel, message);
			pushInputHistory(message);
		}

		public void sendQuietCommandLine(final String serverName, final String channel, final String message) throws RemoteException {
			if (I) Log.i(TAG, "sendQuietCommandLine(): : serverName="+serverName+", channel="+channel+", message="+message);
			ircServerList.sendQuietCommandLine(serverName, channel, message);
		}

		public void sendQuietMessageToChannel(final String serverName, final String channel, final String message) throws RemoteException {
			if (I) Log.i(TAG, "sendQuietMessageToChannel(): serverName="+serverName+", channel="+channel+", message="+message);
			ircServerList.sendQuietMessageToChannel(serverName, channel, message);
			pushInputHistory(message);
		}

		public void sendNoticeToChannel(final String serverName, final String channel, final String message) throws RemoteException {
			if (I) Log.i(TAG, "sendNoticeToChannel(): serverName="+serverName+", channel="+channel+", message="+message);
			ircServerList.sendNoticeToChannel(serverName, channel, message);
			pushInputHistory(message);
		}

		public void echoMessageToChannel(final String serverName, final String channel, final String message) throws RemoteException {
			if (I) Log.i(TAG, "echoMessageToChannel(): serverName="+serverName+", channel="+channel+", message="+message);
			ircServerList.echoMessageToChannel(serverName, channel, message);
		}

		public void connectServer(final String serverName) throws RemoteException {
			if (I) Log.i(TAG, "connectServer(): serverName="+serverName);
			ircServerList.connectServer(serverName);
			ircServerList.saveServerIsConnectedFlag();
		}

		public void disconnectServer(final String serverName) throws RemoteException {
			if (I) Log.i(TAG, "disconnectServer(): serverName="+serverName);
			ircServerList.disconnectServer(serverName);
			ircServerList.saveServerIsConnectedFlag();
		}

		public void closeServer(final String serverName) throws RemoteException {
			if (I) Log.i(TAG, "closeServer(): serverName="+serverName);
			ircServerList.closeServer(serverName);
			ircServerList.saveServerIsConnectedFlag();
		}

		public int connectAuto() throws RemoteException {
			if (I) Log.i(TAG, "connectAuto()");
			int ret = ircServerList.connectAuto();
			ircServerList.saveServerIsConnectedFlag();
			return ret;
		}

		public void disconnectAll() throws RemoteException {
			if (I) Log.i(TAG, "disconnectAll()");
			ircServerList.disconnectAll();
			ircServerList.saveServerIsConnectedFlag();
		}

		public void closeAll() throws RemoteException {
			if (I) Log.i(TAG, "closeAll()");
			ircServerList.closeAll();
			ircServerList.saveServerIsConnectedFlag();
		}

		public synchronized void addNewServer() throws RemoteException {
			if (I) Log.i(TAG, "addNewServer()");
			final int num = ircServerList.getNumServers();
			final IRCServerConfig config = new IRCServerConfig(me);
			config.loadConfig(num);
			final CountDownLatch signal = new CountDownLatch(1);
			final Runnable adder = new Runnable(){ public void run() {
				ircServerList.addServer(config);
				signal.countDown();
			} };
			mHandler.post(adder);
			try {
				signal.await();
			} catch (final InterruptedException e) {
				e.printStackTrace();
			}
		}

		public void reloadServerConfig(final int serverId) throws RemoteException {
			if (I) Log.i(TAG, "reloadServerConfig(): serverId="+serverId);
			ircServerList.reloadServerConfig(serverId);
			final String serverName = ircServerList.getServerName(serverId);
			if (SystemConfig.showServerIcon) {
				showNotification("Update " + serverName, (mDonate ? "AiCiA (DONATED)" : "AiCiA"),
						String.format(me.getString(R.string.ircservice_java_updateserver),serverName),
						R.drawable.icon_disconnect,
						0,
						NOTIFY_SERVER_MESSAGE + ircServerList.getServerId(serverName)
				);
			}
		}

		public synchronized void removeServer(final int serverId) throws RemoteException {
			if (I) Log.i(TAG, "removeServer(): serverId="+serverId);
			final CountDownLatch signal = new CountDownLatch(1);
			final Runnable remover = new Runnable(){ public void run() {
				ircServerList.removeServer(serverId);
				signal.countDown();
			} };
			mHandler.post(remover);
			try {
				signal.await();
			} catch (final InterruptedException e) {
				e.printStackTrace();
			}
		}

		public void setCurrentServerName(final String serverName) throws RemoteException {
			if (I) Log.i(TAG, "setCurrentServerName(): serverName="+serverName);
			ircServerList.setCurrentServerName(serverName);
		}

		public String getCurrentServerName() throws RemoteException {
			if (D) Log.d(TAG, "getCurrentServerName()");
			return ircServerList.getCurrentServerName();
		}

		public void setCurrentChannel(final String serverName, final String channel) throws RemoteException {
			if (I) Log.i(TAG, "setCurrentChannel(): serverName="+serverName+", channel="+channel);
			ircServerList.setCurrentChannel(serverName, channel);
		}

		public String getCurrentChannel(final String serverName) throws RemoteException {
			if (D) Log.d(TAG, "getCurrentChannel(): serverName="+serverName);
			return ircServerList.getCurrentChannel(serverName);
		}

		public String[] getServerList() throws RemoteException {
			if (D) Log.d(TAG, "getServerList()");
			return ircServerList.getServerList();
		}

		public String[] getChannelList(final String serverName) throws RemoteException {
			if (D) Log.d(TAG, "getChannelList(): serverName="+serverName);
			return ircServerList.getChannelList(serverName);
		}

		public boolean[] getChannelUpdatedList(final String serverName) throws RemoteException {
			if (D) Log.d(TAG, "getChannelUpdatedList(): serverName="+serverName);
			final Boolean[] b1 = ircServerList.getChannelUpdatedList(serverName);
			final int num = b1.length;
			final boolean[] b2 = new boolean[num];
			for (int i=0; i<num; i++) {
				b2[i] = b1[i];
			}
			return b2;
		}

		public boolean getChannelUpdated(final String serverName, final String channel) throws RemoteException {
			if (D) Log.d(TAG, "getChannelUpdated(): serverName="+serverName+", channel="+channel);
			return ircServerList.getChannelUpdated(serverName, channel);
		}

		public void clearChannelUpdated(final String serverName, final String channel) throws RemoteException {
			if (I) Log.i(TAG, "clearChannelUpdated(): serverName="+serverName+", channel="+channel);
			ircServerList.clearChannelUpdated(serverName, channel);
		}

		public void clearChannelUpdatedAll() throws RemoteException {
			if (I) Log.i(TAG, "clearChannelUpdatedAll()");
			for (int i=0; i<ircServerList.getNumServers(); i++) {
				final IRCServer ircServer = ircServerList.getServer(i);
				for (int j=0; j<ircServer.getNumChannels(); j++) {
					ircServer.clearChannelUpdated(j);
				}
			}
		}

		public boolean[] getChannelAlertedList(final String serverName) throws RemoteException {
			if (D) Log.d(TAG, "getChannelAlertedList()");
			final Boolean[] b1 = ircServerList.getChannelAlertedList(serverName);
			final int num = b1.length;
			final boolean[] b2 = new boolean[num];
			for (int i=0; i<num; i++) {
				b2[i] = b1[i];
			}
			return b2;
		}

		public boolean getChannelAlerted(final String serverName, final String channel) throws RemoteException {
			if (D) Log.d(TAG, "getChannelAlerted()");
			return ircServerList.getChannelAlerted(serverName, channel);
		}

		public void clearChannelAlerted(final String serverName, final String channel) throws RemoteException {
			if (D) Log.d(TAG, "clearChannelAlerted()");
			ircServerList.clearChannelAlerted(serverName, channel);
		}

		public void clearChannelAlertedAll() throws RemoteException {
			if (D) Log.d(TAG, "clearChannelAlertedAll()");
			for (int i=0; i<ircServerList.getNumServers(); i++) {
				final IRCServer ircServer = ircServerList.getServer(i);
				for (int j=0; j<ircServer.getNumChannels(); j++) {
					ircServer.clearChannelAlerted(j);
				}
			}
		}

		public String[] getUserList(final String serverName, final String channel) throws RemoteException {
			if (D) Log.d(TAG, "getChannelList()");
			return ircServerList.getUserList(serverName, channel);
		}

		public boolean changeNextChannel(final int dir, final int mode) throws RemoteException {
			if (D) Log.d(TAG, "changeNextChannel()");
			final ArrayList<String> flatServerName = new ArrayList<String>();
			final ArrayList<String> flatChannel = new ArrayList<String>();
			final ArrayList<Boolean> flatChannelUpdated = new ArrayList<Boolean>();
			final ArrayList<Boolean> flatChannelAlerted = new ArrayList<Boolean>();
			final int currentId = getFlatServerChannelList(flatServerName, flatChannel, flatChannelUpdated, flatChannelAlerted);
			final int numFlat = flatServerName.size();
			if (numFlat < 1) return false;
			int newId = currentId;
			while (true) {
				if (dir > 0) {
					newId++;
					if (newId >= numFlat) newId = 0;
				} else {
					newId--;
					if (newId < 0) newId = numFlat-1;
				}
				if (newId == currentId) break;
				if (flatChannel.get(newId).equals(IRCMsg.sSystemChannelName)) continue;
				if (mode == 0) break;
				if (flatChannelUpdated.get(newId)) break;
				if (flatChannelAlerted.get(newId)) break;
			}
			if (newId == currentId) return false;
			ircServerList.setCurrentServerName(flatServerName.get(newId));
			ircServerList.setCurrentChannel(flatServerName.get(newId), flatChannel.get(newId));
			return true;
		}

		public void joinChannel(final String serverName, final String channel) throws RemoteException {
			if (D) Log.d(TAG, "joinChannel()");
			ircServerList.joinChannel(serverName, channel);
		}

		public void joinChannels(final String serverName, final String channels) throws RemoteException {
			if (D) Log.d(TAG, "joinChannels()");
			ircServerList.joinChannels(serverName, channels);
		}

		public void partChannel(final String serverName, final String channel) throws RemoteException {
			if (D) Log.d(TAG, "partChannel()");
			ircServerList.joinChannel(serverName, channel);
		}

		public void saveInputBox(final String input, final int selStart, final int selEnd) throws RemoteException {
			if (D) Log.d(TAG, "saveInputBox()");
			mInputBox = new String(input);
			mInputBoxSelStart = selStart;
			mInputBoxSelEnd = selEnd;
		}

		public String loadInputBox() throws RemoteException {
			if (D) Log.d(TAG, "loadInputBox()");
			return new String(mInputBox);
		}

		public int loadInputBoxSelStart() throws RemoteException {
			if (D) Log.d(TAG, "loadInputBoxSelStart()");
			return mInputBoxSelStart;
		}

		public int loadInputBoxSelEnd() throws RemoteException {
			if (D) Log.d(TAG, "loadInputBoxSelEnd()");
			return mInputBoxSelEnd;
		}

		public void receivePong() throws RemoteException {
			if (D) Log.d(TAG, "receivePong()");
		}

		public boolean isTIGMode(final String serverName) throws RemoteException {
			if (D) Log.d(TAG, "isTIGMode()");
			return ircServerList.isTIGMode(serverName);
		}

		public void clearLowMemoryNotification() throws RemoteException {
			if (D) Log.d(TAG, "clearLowMemoryNotice()");
			clearNotification(NOTIFY_LOWMEMORY);
		}

		public String getVersionString() throws RemoteException {
			if (D) Log.d(TAG, "getVersionString()");
			return SystemConfig.getVersionString();
		}

		public boolean isPutOnSublog(final String serverName, final String channel) throws RemoteException {
			if (D) Log.d(TAG, "isPutOnSublog()");
			return ircServerList.isPutOnSublog(serverName, channel);
		}

		public boolean isPutOnSublogAll(final String serverName, final String channel) throws RemoteException {
			if (D) Log.d(TAG, "isPutOnSublogAll()");
			return ircServerList.isPutOnSublogAll(serverName, channel);
		}

		public String getInputHistory(final int pos) throws RemoteException {
			if (D) Log.d(TAG, "getInputHistory()");
			return getInputHistoryMessage(pos);
		}

		public void pushInputHistory(final String message) throws RemoteException {
			if (D) Log.d(TAG, "addInputHistory()");
			pushInputHistoryMessage(message);
		}

		public boolean haveTIGModeInConnectedServers() throws RemoteException {
			if (D) Log.d(TAG, "haveTIGModeInConnectedServers()");
			for (int i=0; i<ircServerList.getNumServers(); i++) {
				final IRCServer ircServer = ircServerList.getServer(i);
				if (ircServer.isTIGMode()) {
					if (ircServer.isConnected()) {
						return true;
					}
					if (ircServer.isTimerEnabled()) {
						return true;
					}
				}
			}
			return false;
		}

	};

	/**
	 * アラート処理
	 */
	private void doAlert() {
		if (D) Log.d(TAG, "doAlert()");
		if (SystemConfig.vibrateOnAlert) {
			new Thread(new Runnable() {	public void run() {
				final Vibrator vib = (Vibrator)getSystemService(VIBRATOR_SERVICE);
				final long[] pat = {0,200,200,200,200,200,200,200};
				vib.vibrate(pat, -1);
			} }).start();
		}
		if (SystemConfig.ringOnAlert) {
			/*
			new Thread(new Runnable() {	public void run() {
				final ToneGenerator tg = new ToneGenerator(
						AudioManager.STREAM_SYSTEM,
						ToneGenerator.MAX_VOLUME);
				tg.startTone(ToneGenerator.TONE_PROP_PROMPT);
				try {
					Thread.sleep(500);
				} catch (final InterruptedException e) {
					//
				}
				tg.stopTone();
			} }).start();
			*/
			try {
				switch (SystemConfig.ringLevel) {
				default:
				case 0:
					mSoundVolume = 0.25F;
					break;
				case 1:
					mSoundVolume = 0.5F;
					break;
				case 2:
					mSoundVolume = 1.0F;
					break;
				}
				if (mSoundId != 0) {
					mSoundPool.stop(mSoundId);
				}
				mSoundId = mSoundPool.play(
						mSound, mSoundVolume, mSoundVolume,
						0, 0, 1
				);
			} catch (final Exception ex) {
				//
			}
		}
	}

	/**
	 * 通知音の読み込み
	 */
	private void loadRingSound() {
		try {
			if (mSoundPool != null) {
				mSoundPool.release();
				mSoundPool = null;
				mSound = 0;
			}
			mSoundPool = new SoundPool(1, AudioManager.STREAM_MUSIC, 0);
			try {
				String path = SystemConfig.getExternalPath();
				path += "ring.ogg";
				mSound = mSoundPool.load(path, 1);
			} catch (final Exception e) {
				//
			}
			if (mSound == 0) {
				mSound = mSoundPool.load(me, R.raw.ring, 1);
			}
		} catch (final Exception ex) {
			//
		}
	}

	/**
	 * 入力メッセージを履歴登録
	 * @param message 入力メッセージ
	 */
	private void pushInputHistoryMessage(String message) {
		if (message == null) return;
		if (message.equals("")) return;
		for (int i=0; i<mInputHistory.size(); i++) {
			if (mInputHistory.get(i).equals(message)) {
				mInputHistory.remove(i);
				break;
			}
		}
		mInputHistory.add(message);
		if (mInputHistory.size() > mInputHistorySize) {
			mInputHistory.remove(0);
		}
	}

	/**
	 * 入力メッセージ履歴を参照
	 * @param pos 位置
	 * @return 履歴内容
	 */
	private String getInputHistoryMessage(int pos) {
		if (pos >= mInputHistory.size()) {
			return null;
		}
		return mInputHistory.get(mInputHistory.size()-pos-1);
	}

	/**
	 * タイマー処理
	 */
	/*
	private boolean mTimerStop = false;
	private static final long mTimerStep = 60000;
	private long mTimerCount = 0;
	private long mTimerNext = 0;
	private static final int TIMER = 1;
	private final Handler mTimerHandler = new Handler() {
		@Override
		public void handleMessage(final Message msg) {
			if (msg.what == TIMER) {
				while (SystemClock.uptimeMillis() > mTimerNext) {
					mTimerNext += mTimerStep;
					mTimerCount++;
					onTimer(mTimerCount);
				}
				if (mTimerStop == false) {
					final Message msg2 = mTimerHandler.obtainMessage(TIMER);
					sendMessageAtTime(msg2, mTimerNext);
				}
			}
		}
	};
	private void startTimer() {
		if (D) Log.d(TAG, "startTimer()");
		mTimerStop = false;
		mTimerNext = SystemClock.uptimeMillis();
		final Message msg = mTimerHandler.obtainMessage(TIMER);
		mTimerHandler.sendMessageAtTime(msg, mTimerNext+mTimerStep);
	}
	private void stopTimer() {
		if (D) Log.d(TAG, "stopTimer()");
		mTimerStop = true;
	}
	private long getTimerCount() {
		return mTimerCount;
	}
	private void onTimer(final long timerCount) {
		if (D) Log.d(TAG, "onTimer()");
		// ActivityMainにKeepalive Pingを送る
		final Intent intent = new Intent(ACTION);
		intent.putExtra("msg", IRCService.PING_PONG);
		intent.putExtra("timer", getTimerCount());
		sendBroadcast(intent);
	}
	 */

	/**
	 * アラーム処理
	 */
	/*
	private void setAlarmManager() {
		if (D) Log.d(TAG, "setAlarmManager()");
		if (mAlarmManager == null) {
			Intent intent = new Intent(IRCService.this, AlarmManagerReceiver.class);
			PendingIntent sender = PendingIntent.getBroadcast(this, 0, intent, 0);
			long firstTime = SystemClock.elapsedRealtime();
			firstTime += 60 * 1000;
			mAlarmManager = (AlarmManager)getSystemService(ALARM_SERVICE);
			mAlarmManager.setRepeating(AlarmManager.ELAPSED_REALTIME_WAKEUP, firstTime, 60 * 1000, sender);
			if (D) Log.d(TAG, "setAlarmManager(): set alarm");
		}
	}

	private void resetAlarmManager() {
		if (D) Log.d(TAG, "resetAlarmManager()");
		if (mAlarmManager != null) {
			Intent intent = new Intent(IRCService.this, AlarmManagerReceiver.class);
			PendingIntent sender = PendingIntent.getBroadcast(this, 0, intent, 0);
			AlarmManager am = (AlarmManager) getSystemService(ALARM_SERVICE);
			am.cancel(sender);
			mAlarmManager = null;
			if (D) Log.d(TAG, "resetAlarmManager(): reset alarm");
		}

	}
	*/

}
